/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"

#include <string.h>
#include <sys/stat.h>
#include <dirent.h>
#include <utime.h>
#include <ctype.h>

int mutt_parse_mailboxes (const char *s, void *data, char *err, size_t errlen)
{
  BUFFY **tmp;
  char buf[_POSIX_PATH_MAX];
  const char *p;
  size_t len;

  do {
    if ((p = strpbrk (s, " \t\n")))
      len = p - s;
    else
      len = strlen (s);

    if (len > sizeof (buf) - 1)
      len = sizeof (buf) - 1;

    memcpy (buf, s, len);
    buf[len] = 0;

    mutt_expand_path (buf, sizeof (buf));
    tmp = &Incoming;
    while (*tmp) /* simple check to avoid duplicates */
    {
      if (strcmp (buf, (*tmp)->path) == 0)
	break;
      tmp = &((*tmp)->next);
    }

    if (!*tmp)
    {
      *tmp = (BUFFY *) safe_malloc (sizeof (BUFFY));
      (*tmp)->path = safe_strdup (buf);
      (*tmp)->new = 0;
      (*tmp)->next = NULL;
    }

    if (p)
      SKIPWS (p);

  } while ((s = p));

  return 0;
}

int mutt_buffy_check (void)
{
  static time_t	last = 0; /* to avoid too many checks */
  static int	ct = 0;
  BUFFY		*tmp = Incoming;
  struct stat	sb;
  time_t t;
  struct dirent *de;
  DIR *dirp;
  char path[_POSIX_PATH_MAX];
  int typ;

  if (tmp == NULL) return 0;
  t = time (NULL);
  if ((t - last) < 3) return ct;
  last = t;
  ct = 0;
  while (tmp)
  {
    tmp->new = 0;
    if ((!Context ||
	(Context && Context->path && strcmp (tmp->path, Context->path))) &&
	stat (tmp->path, &sb) != -1)
    {
      typ = mx_get_magic (tmp->path);

      switch (typ)
      {
	case M_MBOX:
	case M_MMDF:
	  if (sb.st_mtime > sb.st_atime)
	  {
	    ct += 1;
	    tmp->new = 1;	
	  }
	  break;
	case M_MAILDIR:
	  snprintf (path, sizeof (path), "%s/new", tmp->path);
	  if ((dirp = opendir (path)) == NULL) break;
	  while ((de = readdir (dirp)) != NULL)
	  {
	    if (*de->d_name != '.')
	    {
	      /* one new message is enough */
	      ct++;
	      tmp->new = 1;
	      break;
	    }
	  }
	  closedir (dirp);
	  break;
	case M_MH:
	  if ((dirp = opendir (tmp->path)) == NULL) break;
	  while ((de = readdir (dirp)) != NULL)
	  {
	    char *p;

             p = de->d_name;
             while (*p)
	     {
	       if (!isdigit (*p)) break;
	       p++;
	     }
	     if (*p) continue;
	     snprintf (path, sizeof (path), "%s/%s", tmp->path, de->d_name);
	     if (stat (path, &sb) != -1 && sb.st_mtime > sb.st_atime)
	     {
	       ct++;
	       tmp->new = 1;
	       break;
	     }
	  }
	  closedir (dirp);
	  break;
      }
    }
    tmp = tmp->next;
  }
  return (ct);
}

/*
 * mutt_buffy() -- incoming folders completion routine
 *
 * given a folder name, this routine gives the next incoming folder with new
 * new mail.
 */
void mutt_buffy (char *s)
{
  int	count = 0;
  BUFFY	*tmp = Incoming;

  mutt_expand_path (s, _POSIX_PATH_MAX);
  switch (count = mutt_buffy_check ())
  {
    case 0 :
      s = '\0';
      break;
    case 1 :
      while (tmp->new == 0) tmp = tmp->next;
      strcpy (s, tmp->path);
      mutt_pretty_mailbox (s);
      break;
    default :
      FOREVER
      {
	if (strcmp (s, tmp->path) == 0)
	  count = 0;
	else if (count == 0 && tmp->new != 0)
	  break;
	tmp = tmp->next;
	if (tmp == NULL)
	{
	  tmp = Incoming;
	  count = 0;
	}
      }
      strcpy (s, tmp->path);
      mutt_pretty_mailbox (s);
      break;
  }
}
