/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "send.h"
#include "rfc1524.h"

#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>

#define CHECK_COUNT if (AttachLen == 0) { mutt_error ("There are no attachments."); break; }

#define HDR_TO 1
#define HDR_CC 2
#define HDR_BCC 3
#define HDR_SUBJECT 4
#define HDR_REPLYTO 5
#define HDR_FCC 6
#define HDR_PGP 7

/* where to start printing the attachments */
#define HDR_ATTACH 10

#define HDR_XOFFSET 10

static BODY **Attach = NULL;
static short AttachLen = 0;
static short AttachMax = 0;

static void gen_attach_list (BODY *m)
{
  while (m)
  {
    if (m->type == TYPEMULTIPART)
    {
      gen_attach_list (m->parts);
    }
    else
    {
      if (AttachLen == AttachMax)
	safe_realloc ((void **) &Attach, sizeof (BODY *) * (AttachMax += 5));

      Attach[AttachLen++] = m;

      /* Hmm, this was skipped before because m->parts was NULL, as we
       * weren't parsing forwarded messages.  This now causes a core
       * dump because there is no separate file for each part of the
       * forward, yet.
       */
#if 0
      if (m->type == TYPEMESSAGE &&
	  (!strcasecmp (m->subtype, "rfc822") || !strcasecmp (m->subtype, "news")) &&
	  m->parts && is_multipart (m->parts))
      {
	gen_attach_list (m->parts);
      }
#endif
    }
    m = m->next;
  }
}

void snd_entry (char *b, size_t blen, MUTTMENU *menu, int num)
{
  char t[SHORT_STRING], size[SHORT_STRING];
  char tmp[_POSIX_PATH_MAX];
  BODY *m = Attach[num];
  struct stat finfo;

  stat (m->filename, &finfo);
  mutt_pretty_size (size, sizeof (size), finfo.st_size);

  strfcpy (tmp, m->filename, sizeof (tmp));
  mutt_pretty_mailbox (tmp);

  snprintf (t, sizeof (t), "[%.7s/%.10s, %.6s, %s]",
	    TYPE (m->type), m->subtype, ENCODING (m->encoding), size);

  snprintf (b, blen, "%c %-2d %s %s %s",
	    m->unlink ? '-' : ' ', num + 1, tmp, t,
	    m->description ? m->description : "<no description>");
}


static void draw_envelope (HEADER *msg, char *fcc)
{
  char buf[STRING];

  mvaddstr (HDR_TO, 0,      "      To: ");
  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), msg->env->to);
  printw ("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buf);

  mvaddstr (HDR_CC, 0,      "      Cc: ");
  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), msg->env->cc);
  printw ("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buf);

  mvaddstr (HDR_BCC, 0,     "     Bcc: ");
  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), msg->env->bcc);
  printw ("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buf);

  mvaddstr (HDR_SUBJECT, 0, " Subject: ");
  if (msg->env->subject)
    printw ("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, msg->env->subject);
  else
    clrtoeol ();

  mvaddstr (HDR_REPLYTO, 0, "Reply-To: ");
  if (msg->env->reply_to)
  {
    buf[0] = 0;
    rfc822_write_address (buf, sizeof (buf), msg->env->reply_to);
    printw ("%-*.*s", COLS - HDR_XOFFSET, COLS - HDR_XOFFSET, buf);
  }
  else
    clrtoeol ();

  mvaddstr (HDR_FCC, 0,     "     Fcc: ");
  addstr (fcc);


  mvaddstr (HDR_ATTACH-1, 0, "===== Attachments =====");
}

static int edit_address_list (int line, ENVELOPE *env)
{
  char buf[HUGE_STRING] = ""; /* needs to be large for alias expansion */
  ADDRESS **addr;
  char *prompt;

  switch (line)
  {
    case HDR_TO:

      prompt = "To: ";
      addr = &env->to;
      break;

    case HDR_CC:

      prompt = "Cc: ";
      addr = &env->cc;
      break;

    case HDR_BCC:

      prompt = "Bcc: ";
      addr = &env->bcc;
      break;

    case HDR_REPLYTO:

      prompt = "Reply-To: ";
      addr = &env->reply_to;
      break;

    default:

      return 0;
  }

  rfc822_write_address (buf, sizeof (buf), *addr);
  if (ci_get_field (prompt, buf, sizeof (buf), M_ALIAS) != 0)
    return 0;

  mutt_free_address (addr);
  mutt_parse_adrlist (addr, buf, "@");
  *addr = mutt_expand_aliases (*addr);

  if (option (OPTNEEDREDRAW))
  {
    unset_option (OPTNEEDREDRAW);
    return (REDRAW_FULL);
  }

  /* redraw the expanded list so the user can see the result */
  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), *addr);
  move (line, HDR_XOFFSET);
  printw ("%-*.*s", COLS - HDR_XOFFSET, COLS - HDR_XOFFSET, buf);

  return 0;
}

/* return values:
 *
 * 1	message should be postponed
 * 0	normal exit
 * -1	abort message
 */
int mutt_send_menu (HEADER *msg,   /* structure for new message */
		    char *fcc,     /* where to save a copy of the message */
		    size_t fcclen,
		    HEADER *cur)   /* current message */
{
  char helpstr[SHORT_STRING];
  char buf[LONG_STRING];
  MUTTMENU *menu;
  int i;
  int r = -1; /* return value */
  int loop = 1;
  int fccSet = 0; /* has the user edited the Fcc: field ? */

  gen_attach_list (msg->content);

  menu = mutt_new_menu ();
  menu->menu = MENU_COMPOSE;
  menu->offset = HDR_ATTACH;
  menu->max = AttachLen;
  menu->make_entry = snd_entry;
  menu->title = "Compose";

  helpstr[0] = 0;
  mutt_make_help (buf, sizeof (buf), "Send  ", MENU_COMPOSE, OP_COMPOSE_SEND_MESSAGE);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Abort  ", MENU_COMPOSE, OP_EXIT);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "To  ", MENU_COMPOSE, OP_COMPOSE_EDIT_TO);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "CC  ", MENU_COMPOSE, OP_COMPOSE_EDIT_CC);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Subj  ", MENU_COMPOSE, OP_COMPOSE_EDIT_SUBJECT);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Attach  ", MENU_COMPOSE, OP_COMPOSE_ATTACH_FILE);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Descrip  ", MENU_COMPOSE, OP_COMPOSE_EDIT_DESCRIPTION);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Help", MENU_COMPOSE, OP_HELP);
  strcat (helpstr, buf);
  menu->help = helpstr;

  while (loop)
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_REDRAW:

	draw_envelope (msg, fcc);
	menu->offset = HDR_ATTACH;
	menu->pagelen = LINES - 12;
	break;

      case OP_COMPOSE_EDIT_TO:

	menu->redraw = edit_address_list (HDR_TO, msg->env);
	break;

      case OP_COMPOSE_EDIT_BCC:

	menu->redraw = edit_address_list (HDR_BCC, msg->env);
	break;

      case OP_COMPOSE_EDIT_CC:

	menu->redraw = edit_address_list (HDR_CC, msg->env);
	break;

      case OP_COMPOSE_EDIT_SUBJECT:

	if (msg->env->subject)
	  strfcpy (buf, msg->env->subject, sizeof (buf));
	else
	  buf[0] = 0;
	if (ci_get_field ("Subject: ", buf, sizeof (buf), 0) == 0)
	{
	  safe_free ((void **) &msg->env->subject);
	  msg->env->subject = safe_strdup (buf);
	  move (HDR_SUBJECT, HDR_XOFFSET);
	  clrtoeol ();
	  if (msg->env->subject)
	    printw ("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, msg->env->subject);
	}
	break;

      case OP_COMPOSE_EDIT_REPLY_TO:

	menu->redraw = edit_address_list (HDR_REPLYTO, msg->env);
	break;

      case OP_COMPOSE_EDIT_FCC:

	buf[0] = 0;
	strfcpy (buf, fcc, sizeof (buf));
	if (ci_get_field ("Fcc: ", buf, sizeof (buf), M_FILE | M_CLEAR) == 0)
	{
	  strfcpy (fcc, buf, _POSIX_PATH_MAX);
	  mutt_pretty_mailbox (fcc);
	  mvprintw (HDR_FCC, HDR_XOFFSET, "%-*.*s",
		    COLS - HDR_XOFFSET, COLS - HDR_XOFFSET, fcc);
	  fccSet = 1;
	}
	MAYBE_REDRAW (menu->redraw);
	break;

      case OP_COMPOSE_EDIT_MESSAGE:

	if (strcmp ("builtin", Editor) == 0)
	{
	  mutt_builtin_editor (msg->content->filename, &msg->env, cur);
	  menu->redraw = REDRAW_FULL;
	}
	else if (option (OPTEDITHDRS))
	{
	  mutt_edit_headers (Editor, msg->content->filename, &msg->env, fcc, fcclen);
	  menu->redraw = REDRAW_FULL; /* recipients could have changed */
	}
	else
	{
	  mutt_edit_file (Editor, msg->content->filename);
	  menu->redraw = REDRAW_CURRENT;
	}
	mutt_update_encoding (msg->content);
	break;

      case OP_COMPOSE_EDIT_HEADERS:

	mutt_edit_headers (strcmp ("builtin", Editor) == 0 ? Visual : Editor,
			   msg->content->filename, &msg->env, fcc, fcclen);
        mutt_update_encoding (msg->content);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_COMPOSE_ATTACH_FILE:

	if (mutt_enter_fname ("Attach file: ", buf, sizeof (buf), &menu->redraw) == -1)
	  break;
	if (!buf[0])
	  continue;
	mutt_expand_path (buf, sizeof (buf));

	/* check to make sure the file exists and is readable*/
	if (access (buf, R_OK) == -1)
	{
	  mutt_perror (buf);
	  break;
	}

	if (AttachLen == AttachMax)
	  safe_realloc ((void **) &Attach, sizeof (BODY *) * (AttachMax += 5));

	if ((Attach[AttachLen] = mutt_make_attach (buf)) != NULL)
	{
	  menu->current = AttachLen++;
	  menu->max = AttachLen;
	  menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
	}
	else
	  mutt_error ("Unable to attach file!");
	break;

      case OP_DELETE:

	CHECK_COUNT;

	if (menu->current != 0 || menu->max != 1)
	{
	  Attach[menu->current]->next = NULL;
	  mutt_free_body (&Attach[menu->current]);
	  for (i = menu->current; i < AttachLen - 1; i++)
	    Attach[i] = Attach[i+1];
	  menu->max = --AttachLen;
	  if (AttachLen)
	  {
	    if (menu->current > AttachLen - 1)
	      menu->current = AttachLen - 1;
	  }
	  else
	    menu->current = 0;

	  if (menu->current == 0)
	    msg->content = Attach[0];

	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
	}
	else
	  mutt_error ("You may not delete the only attachment.");
	break;

      case OP_COMPOSE_EDIT_DESCRIPTION:

	CHECK_COUNT;
	strfcpy (buf,
		 Attach[menu->current]->description ? Attach[menu->current]->description : "",
		 sizeof (buf));
	if (ci_get_field ("Description: ", buf, sizeof (buf), 0) == 0)
	{
	  safe_free ((void **) &Attach[menu->current]->description);
	  Attach[menu->current]->description = safe_strdup (buf);
	  menu->redraw = REDRAW_CURRENT;
	}
	break;

      case OP_COMPOSE_EDIT_TYPE:

	CHECK_COUNT;
	snprintf (buf, sizeof (buf), "%s/%s", TYPE (Attach[menu->current]->type),
		  Attach[menu->current]->subtype);
	if (ci_get_field ("Content-Type: ", buf, sizeof (buf), 0) == 0 && buf[0])
	{
	  char *p = strchr (buf, '/');

	  if (p)
	  {
	    *p++ = 0;
	    if ((i = mutt_check_mime_type (buf)) != TYPEOTHER)
	    {
	      Attach[menu->current]->type = i;
	      safe_free ((void **)&Attach[menu->current]->subtype);
	      Attach[menu->current]->subtype = safe_strdup (p);
	      menu->redraw = REDRAW_CURRENT;
	    }
	  }
	}
	break;

      case OP_COMPOSE_EDIT_ENCODING:

	CHECK_COUNT;
	strfcpy (buf, ENCODING (Attach[menu->current]->encoding), sizeof (buf));
	if (ci_get_field ("Content-Transfer-Encoding: ", buf, sizeof (buf), 0) == 0 && buf[0])
	{
	  if ((i = mutt_check_encoding (buf)) != ENCOTHER)
	  {
	    Attach[menu->current]->encoding = i;
	    menu->redraw = REDRAW_CURRENT;
	  }
	  else
	    mutt_error ("Invalid encoding.");
	}
	break;

      case OP_COMPOSE_SEND_MESSAGE:

	if (!fccSet && *fcc)
	{
	  if ((i = query_quadoption (OPT_COPY, "Save a copy of this message?")) == -1)
	    break;
	  else if (i == M_NO)
	    *fcc = 0;
	}

	loop = 0;
	r = 0;
	break;

      case OP_COMPOSE_EDIT_FILE:

	CHECK_COUNT;
	mutt_edit_file (strcmp ("builtin", Editor) == 0 ? Visual : Editor,
			Attach[menu->current]->filename);
	mutt_update_encoding (Attach[menu->current]);
	menu->redraw = REDRAW_CURRENT;
	break;

      case OP_COMPOSE_TOGGLE_UNLINK:

	CHECK_COUNT;
	Attach[menu->current]->unlink = !Attach[menu->current]->unlink;
	menu->redraw = REDRAW_CURRENT;
	break;

      case OP_COMPOSE_RENAME_FILE:

	strfcpy (buf, Attach[menu->current]->filename, sizeof (buf));
	mutt_pretty_mailbox (buf);
	if (ci_get_field ("Rename to: ", buf, sizeof (buf), M_FILE) == 0 && buf[0])
	{
	  mutt_expand_path (buf, sizeof (buf));
	  if (mutt_rename_file (Attach[menu->current]->filename, buf))
	  {
	    safe_free ((void **) &Attach[menu->current]->filename);
	    Attach[menu->current]->filename = safe_strdup (buf);
	    menu->redraw = REDRAW_CURRENT;
	  }
	}
	break;

      case OP_COMPOSE_NEW_MIME:

	{
	  char type[STRING];
	  char *p;
	  int itype;
	  FILE *fp;

	  CLEARLINE (LINES-1);
	  buf[0] = 0;
	  if (ci_get_field ("New file: ", buf, sizeof (buf), M_FILE) != 0 || !buf[0])
	    continue;
	  mutt_expand_path (buf, sizeof (buf));
	  /* Call to lookup_mime_type () ?  maybe later */
	  type[0] = 0;
	  if (ci_get_field ("Content-Type: ", type, sizeof (type), 0) != 0 || !type[0])
	    continue;

	  if (!(p = strchr (type, '/')))
	  {
	    mutt_error ("Content-Type is of the form base/sub");
	    continue;
	  }
	  *p++ = 0;
	  if ((itype = mutt_check_mime_type (type)) == TYPEOTHER)
	  {
	    mutt_error ("Unknown Content-Type %s",buf);
	    continue;
	  }
	  if (AttachLen == AttachMax)
	    safe_realloc ((void **) &Attach, sizeof (BODY *) * (AttachMax += 5));

	  /* Touch the file */
	  if (!(fp = fopen (buf,"w")))
	  {
	    mutt_error ("Can't create file %s",buf);
	    continue;
	  }
	  fclose (fp);

	  if ((Attach[AttachLen] = mutt_make_attach (buf)) == NULL)
	  {
	    mutt_error ("What we have here is a failure to make an attachment");
	    continue;
	  }

	  menu->current = AttachLen++;
	  menu->max = AttachLen;
	  Attach[menu->current]->type = itype;
	  safe_free ((void **) &Attach[menu->current]->subtype);
	  Attach[menu->current]->subtype = safe_strdup (p);
	  Attach[menu->current]->unlink = 1;
	  menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
	}

      case OP_COMPOSE_EDIT_MIME:

	CHECK_COUNT;
	if (mutt_edit_attachment (Attach[menu->current], 0))
	{
	  mutt_update_encoding (Attach[menu->current]);
	  menu->redraw = REDRAW_FULL;
	}
	break;

      case OP_COMPOSE_VIEW_ATTACH:

	CHECK_COUNT;
	if (mutt_view_attachment (Attach[menu->current], 0))
	{
	  menu->redraw = REDRAW_FULL;
	  mutt_clear_error ();
	}
	break;

      case OP_SAVE:

	CHECK_COUNT;
	buf[0] = 0;
	if (ci_get_field ("Copy to file: ", buf, sizeof (buf), M_FILE) != 0 || !buf[0])
	  break;
	mutt_expand_path (buf, sizeof (buf));
	if (mutt_save_attachment (NULL, Attach[menu->current], buf) == 0)
	  mutt_message ("Attachment copied.");
	break;

      case OP_PRINT:

	CHECK_COUNT;
	if (query_quadoption (OPT_PRINT, "Print Attachment?") == M_YES)
	{
	  if (mutt_print_attachment (NULL, Attach[menu->current]))
	    menu->redraw = REDRAW_FULL;
	}
	break;

      case OP_PIPE:

	CHECK_COUNT;
	buf[0] = 0;
	if (ci_get_field ("Filter through: ", buf, sizeof (buf), 0) != 0 || !buf[0])
	  break;
	mutt_expand_path (buf, sizeof (buf));

	{
	  char warning[STRING];
	  char outfile[_POSIX_PATH_MAX];

	  snprintf (warning, sizeof (warning),
		    "WARNING! You are about to overwrite %s, continue?",
		    Attach[menu->current]->filename);
	  if (mutt_yesorno (warning, M_NO) == M_YES)
	  {
	    mutt_mktemp (outfile);
	    if (mutt_pipe_attachment (NULL, Attach[menu->current], buf, outfile))
	    {
	      mutt_unlink (Attach[menu->current]->filename);
	      mutt_rename_file (outfile, Attach[menu->current]->filename);
	      menu->redraw = REDRAW_CURRENT;
	      mutt_update_encoding (Attach[menu->current]);
	      mutt_message ("Attachment Filtered");
	    }
	  }
	}
	break;

      case OP_EXIT:

	i = query_quadoption (OPT_POSTPONE, "Postpone this message?");
	if (i == M_NO)
	{
	  for (i = 0; i < AttachLen; i++)
	  {
	    Attach[i]->next = NULL; /* avoid freeing other attachments */
	    mutt_free_body (&Attach[i]);
	  }
	  AttachLen = 0;
	  r = -1;
	  loop = 0;
	  break;
	}
	else if (i == -1)
	  break; /* abort */

	/* fall through to postpone! */

      case OP_COMPOSE_POSTPONE_MESSAGE:

	loop = 0;
	r = 1;
	break;

#ifdef ISPELL
      case OP_COMPOSE_ISPELL:

	mutt_endwin (NULL);
	snprintf (buf, sizeof (buf), "%s -x %s", Ispell, msg->content->filename);
	mutt_system (buf);
	break;
#endif /* ISPELL */

    }
  }

  mutt_menuDestroy (&menu);

  if (AttachLen)
  {
    for (i = 0; i < AttachLen - 1; i++)
      Attach[i]->next = Attach[i+1];
    Attach[i]->next = NULL;
    msg->content = Attach[0];
  }
  else
    msg->content = NULL;

  safe_free ((void **) &Attach);
  AttachLen = 0;
  AttachMax = 0;

  return (r);
}
