/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"

void mutt_set_flag (CONTEXT *ctx, int msgno, int flag, int bf)
{
  HEADER *h = ctx->hdrs[msgno];

  if (ctx->readonly && flag != M_TAG)
    return; /* don't modify anything if we are read-only */

  switch (flag)
  {
    case M_DELETE:

      if (bf)
      {
	if (!h->deleted)
	{
	  h->deleted = 1;
	  ctx->deleted++;
	}
      }
      else if (h->deleted)
      {
	h->deleted = 0;
	ctx->deleted--;
      }
      break;

    case M_NEW:

      if (bf)
      {
	if (h->read || h->old)
	{
	  h->read = 0;
	  h->old = 0;
	  ctx->new++;
	  h->changed = 1;
	  ctx->changed = 1;
	}
      }
      else if (!h->read)
      {
	if (!h->old)
	  ctx->new--;
	h->read = 1;
	h->changed = 1;
	ctx->changed = 1;
      }
      break;

    case M_OLD:
      if (bf)
      {
	if (!h->old)
	{
	  h->old = 1;
	  if (!h->read)
	    ctx->new--;
	  h->changed = 1;
	  ctx->changed = 1;
	}
      }
      else if (h->old)
      {
	h->old = 0;
	if (!h->read)
	  ctx->new++;
	h->changed = 1;
	ctx->changed = 1;
      }
      break;

    case M_READ:

      if (bf)
      {
	if (!h->read)
	{
	  h->read = 1;
	  if (!h->old)
	    ctx->new--;
	  h->changed = 1;
	  ctx->changed = 1;
	}
      }
      else if (h->read)
      {
	h->read = 0;
	if (!h->old)
	  ctx->new++;
	h->changed = 1;
	ctx->changed = 1;
      }
      break;

    case M_REPLIED:

      if (bf)
      {
	if (!h->replied)
	{
	  h->replied = 1;
	  if (!h->read)
	  {
	    h->read = 1;
	    if (!h->old)
	      ctx->new--;
	  }
	  h->changed = 1;
	  ctx->changed = 1;
	}
      }
      else if (h->replied)
      {
	h->replied = 0;
	h->changed = 1;
	ctx->changed = 1;
      }
      break;

    case M_FLAG:

      if (bf)
      {
	if (!h->flagged)
	{
	  h->flagged = bf;
	  ctx->flagged++;
	  h->changed = 1;
	  ctx->changed = 1;
	}
      }
      else if (h->flagged)
      {
	h->flagged = 0;
	ctx->flagged--;
	h->changed = 1;
	ctx->changed = 1;
      }
      break;

    case M_TAG:

      if (bf)
      {
	if (!h->tagged)
	{
	  h->tagged = 1;
	  ctx->tagged++;
	}
      }
      else if (h->tagged)
      {
	h->tagged = 0;
	ctx->tagged--;
      }
      break;
  }
}

void mutt_tag_set_flag (int flag, int bf)
{
  int j;

  for (j = 0; j < Context->vcount; j++)
    if (Context->hdrs[Context->v2r[j]]->tagged)
      mutt_set_flag (Context, Context->v2r[j], flag, bf);
}

static void thread_set_flag_r (HEADER *h, int flag, int bf)
{
  for (; h; h = h->next)
  {
    mutt_set_flag (Context, h->msgno, flag, bf);
    if (h->child)
      thread_set_flag_r (h->child, flag, bf);
  }
}

void mutt_thread_set_flag (HEADER *cur, int flag, int bf)
{
  while (cur->parent)
    cur = cur->parent;
  mutt_set_flag (Context, cur->msgno, flag, bf);
  thread_set_flag_r (cur->child, flag, bf);
}

int mutt_change_flag (HEADER *h, int bf)
{
  int i, flag;

  mvprintw (LINES - 1, 0, "%s flag? (D/N/O/r/*/!): ", bf ? "Set" : "Clear");
  clrtoeol ();

  if ((i = ci_getch ()) == ERR)
  {
    CLEARLINE (LINES-1);
    return (-1);
  }

  CLEARLINE (LINES-1);

  switch (i)
  {
    case 'd':
    case 'D':
      flag = M_DELETE;
      break;

    case 'N':
    case 'n':
      flag = M_NEW;
      break;

    case 'o':
    case 'O':
      if (h)
	mutt_set_flag (Context, h->msgno, M_READ, !bf);
      else
	mutt_tag_set_flag (M_READ, !bf);
      flag = M_OLD;
      break;

    case 'r':
    case 'R':
      flag = M_REPLIED;
      break;

    case '*':
      flag = M_TAG;
      break;

    case '!':
      flag = M_FLAG;
      break;

    default:
      BEEP ();
      return (-1);
  }

  if (h)
    mutt_set_flag (Context, h->msgno, flag, bf);
  else
    mutt_tag_set_flag (flag, bf);

  return 0;
}
