/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

/*
 * This file contains the structures needed to parse ``bind'' commands, as
 * well as the default bindings for each menu.
 *
 * Notes:
 *
 * - If you want to bind \n or \r, use M_ENTER_S so that it will work
 * correctly under both ncurses and S-Lang
 *
 * - If you need to bind a control char, use the octal value because the \cX
 * construct does not work at this level.
 *
 */

struct binding_t OpGeneric[] = {
  { "select-entry",	OP_GENERIC_SELECT_ENTRY,M_ENTER_S },
  { "next-entry",	OP_NEXT_ENTRY,		"j" },
  { "previous-entry",	OP_PREV_ENTRY,		"k" },
  { "last-entry",	OP_LAST_ENTRY,		"*" },
  { "first-entry",	OP_FIRST_ENTRY,		"=" },
  { "next-page",	OP_NEXT_PAGE,		"z" },
  { "previous-page",	OP_PREV_PAGE,		"Z" },
  { "help",		OP_HELP,		"?" },
  { "refresh",		OP_REDRAW,		"\014" },
  { "top-page",		OP_TOP_PAGE,		"H" },
  { "middle-page",	OP_MIDDLE_PAGE,		"M" },
  { "bottom-page",	OP_BOTTOM_PAGE,		"L" },
  { "search-next",	OP_SEARCH_NEXT,		"n" },
  { "exit",		OP_EXIT,		"q" },
  { "search",		OP_SEARCH,		"/" },
  { "search-reverse",	OP_SEARCH_REVERSE,	"\033/" },
  { "enter-command",	OP_ENTER_COMMAND,	":" },
  { "next-line",	OP_NEXT_LINE,		">" },
  { "previous-line",	OP_PREV_LINE,		"<" },
  { "jump",		OP_JUMP,		NULL },
  { NULL,		0,			NULL }
};

struct binding_t OpMain[] = {
  { "create-alias",		OP_CREATE_ALIAS,		"a" },
  { "bounce-message",		OP_BOUNCE_MESSAGE,		"b" },
  { "bottom-page",		OP_BOTTOM_PAGE,			"B" },
  { "browse-url",		OP_BROWSE_URL,			"\002" },
  { "change-folder",		OP_MAIN_CHANGE_FOLDER,		"c" },
  { "change-folder-readonly",	OP_MAIN_CHANGE_FOLDER_READONLY,	"\033c" },
  { "copy-message",		OP_COPY_MESSAGE,		"C" },
  { "delete-message",		OP_DELETE,			"d" },
  { "delete-pattern",		OP_MAIN_DELETE_PATTERN,		"D" },
  { "delete-thread",		OP_DELETE_THREAD,		"\004" },
  { "forward-message",		OP_FORWARD_MESSAGE,		"f" },
  { "flag-message",		OP_FLAG_MESSAGE,		"F" },
  { "group-reply",		OP_GROUP_REPLY,			"g" },
  { "fetch-mail",		OP_MAIN_FETCH_MAIL,		"G" },
  { "display-headers",		OP_DISPLAY_HEADERS,		"h" },
  { "top-page",			OP_TOP_PAGE,			"H" },
  { "next-undeleted",		OP_MAIN_NEXT_UNDELETED,		"j" },
  { "next-message",		OP_NEXT_ENTRY,			"J" },
  { "previous-undeleted",	OP_MAIN_PREV_UNDELETED,		"k" },
  { "previous-message",		OP_PREV_ENTRY,			"K" },
  { "limit",			OP_MAIN_LIMIT,			"l" },
  { "redraw-screen",		OP_REDRAW,			"\014" },
  { "list-reply",		OP_LIST_REPLY,			"L" },
  { "mail",			OP_MAIL,			"m" },
  { "middle-page",		OP_MIDDLE_PAGE,			"M" },
  { "toggle-new",		OP_TOGGLE_NEW,			"N" },
  { "toggle-write",		OP_TOGGLE_WRITE,		"%" },
  { "search-next",		OP_SEARCH_NEXT,			"n" },
  { "next-thread",		OP_MAIN_NEXT_THREAD,		"\016" },
  { "quit",			OP_QUIT,			"q" },
  { "reply",			OP_REPLY,			"r" },
  { "sort-mailbox",		OP_SORT,			"o" },
  { "sort-reverse",		OP_SORT_REVERSE,		"O" },
  { "print-message",		OP_PRINT,			"p" },
  { "previous-thread",		OP_MAIN_PREV_THREAD,		"\020" },
  { "recall-message",		OP_RECALL_MESSAGE,		"R" },
  { "read-thread",		OP_MAIN_READ_THREAD,		"\022" },
  { "save-message",		OP_SAVE,			"s" },
  { "search-body",		OP_MAIN_SEARCH_BODY,		"S" },
  { "tag-message",		OP_TAG_MESSAGE,			"t" },
  { "tag-pattern",		OP_MAIN_TAG_PATTERN,		"T" },
  { "untag-pattern",		OP_MAIN_UNTAG_PATTERN,		"\024" },
  { "undelete-message",		OP_UNDELETE,			"u" },
  { "undelete-pattern",		OP_MAIN_UNDELETE_PATTERN,	"U"},
  { "undelete-thread",		OP_UNDELETE_THREAD,		"\025" },
  { "view-attachments",		OP_VIEW_ATTACHMENTS,		"v" },
  { "show-version",		OP_VERSION,			"V" },
  { "set-flag",			OP_MAIN_SET_FLAG,		"w" },
  { "clear-flag",		OP_MAIN_CLEAR_FLAG,		"W" },
  { "exit",			OP_MAIN_EXIT,			"x" },
  { "next-page",		OP_NEXT_PAGE,			"z" },
  { "previous-page",		OP_PREV_PAGE,			"Z" },
  { "display-message",		OP_DISPLAY_MESSAGE,		M_ENTER_S },
  { "first-message",		OP_FIRST_ENTRY,			"=" },
  { "last-message",		OP_LAST_ENTRY,			"*" },
  { "enter-command",		OP_ENTER_COMMAND,		":" },
  { "sync-mailbox",		OP_MAIN_SYNC_FOLDER,		"$" },
  { "tag-prefix",		OP_MAIN_TAG_PREFIX,		";" },
  { "search",			OP_SEARCH,			"/" },
  { "search-reverse",		OP_SEARCH_REVERSE,		"\033/" },
  { "help",			OP_HELP,			"?" },
  { "shell-escape",		OP_SHELL_ESCAPE,		"!" },
  { "display-address",		OP_DISPLAY_ADDRESS,		"@" },
  { "pipe-message",		OP_PIPE,			"|" },
  { "next-new",			OP_MAIN_NEXT_NEW,		"\t" },
  { "previous-new",		OP_MAIN_PREV_NEW,		"\033\t" },
  { "next-line",        	OP_NEXT_LINE,           	">" },
  { "previous-line",    	OP_PREV_LINE,           	"<" },
  { "jump",			OP_JUMP,			NULL },
  { "next-unread",		OP_MAIN_NEXT_UNREAD,		NULL },
  { "previous-unread",		OP_MAIN_PREV_UNREAD,		NULL },
  { NULL,			0,				NULL }
};

struct binding_t OpPager[] = {
  { "create-alias",	OP_CREATE_ALIAS,		"a" },
  { "bounce-message",	OP_BOUNCE_MESSAGE,		"b" },
  { "browse-url",	OP_BROWSE_URL,			"\002" },
  { "change-folder",	OP_MAIN_CHANGE_FOLDER,		"c" },
  { "change-folder-readonly",	OP_MAIN_CHANGE_FOLDER_READONLY,	"\033c" },
  { "copy-message",	OP_COPY_MESSAGE,		"C" },
  { "delete-message",	OP_DELETE,			"d" },
  { "delete-thread",	OP_DELETE_THREAD,		"\004" },
  { "forward-message",	OP_FORWARD_MESSAGE,		"f" },
  { "flag-message",	OP_FLAG_MESSAGE,		"F" },
  { "group-reply",	OP_GROUP_REPLY,			"g" },
  { "display-headers",	OP_DISPLAY_HEADERS,		"h" },
  { "exit",		OP_PAGER_EXIT,			"i" },
  { "next-undeleted",	OP_MAIN_NEXT_UNDELETED,		"j" },
  { "next-message",	OP_NEXT_ENTRY,			"J" },
  { "previous-undeleted",OP_MAIN_PREV_UNDELETED,	"k" },
  { "previous-message",	OP_PREV_ENTRY,			"K" },
  { "list-reply",	OP_LIST_REPLY,			"L" },
  { "redraw-screen",	OP_REDRAW,			"\014" },
  { "mail",		OP_MAIL,			"m" },
  { "search-next",      OP_SEARCH_NEXT,         	"n" },
  { "mark-as-new",	OP_TOGGLE_NEW,			"N" },
  { "next-thread",	OP_MAIN_NEXT_THREAD,		"\016" },
  { "print-message",	OP_PRINT,			"p" },
  { "previous-thread",	OP_MAIN_PREV_THREAD,		"\020" },
  { "quit",		OP_QUIT,			"Q" },
  { "reply",		OP_REPLY,			"r" },
  { "recall-message",	OP_RECALL_MESSAGE,		"R" },
  { "read-thread",	OP_MAIN_READ_THREAD,		"\022" },
  { "save-message",	OP_SAVE,			"s" },
  { "tag-message",	OP_TAG_MESSAGE,			"t" },
  { "toggle-quoted",	OP_PAGER_HIDE_QUOTED,		"T" },
  { "undelete-message",	OP_UNDELETE,			"u" },
  { "undelete-thread",	OP_UNDELETE_THREAD,		"\025" },
  { "view-attachments",	OP_VIEW_ATTACHMENTS,		"v" },
  { "show-version",	OP_VERSION,			"V" },
  { "search-toggle",	OP_SEARCH_TOGGLE,		"\\" },
  { "display-address",	OP_DISPLAY_ADDRESS,		"@" },
  { "next-new",		OP_MAIN_NEXT_NEW,		"\t" },
  { "pipe-message",	OP_PIPE,			"|" },
  { "help",		OP_HELP,			"?" },
  { "next-page",	OP_NEXT_PAGE,			" " },
  { "previous-page",	OP_PREV_PAGE,			"-" },
  { "top",		OP_PAGER_TOP,			"^" },
  { "bottom",		OP_PAGER_BOTTOM,		"$" },
  { "shell-escape",	OP_SHELL_ESCAPE,		"!" },
  { "search",		OP_SEARCH,			"/" },
  { "search-reverse",	OP_SEARCH_REVERSE,		"\033/" },
  { "enter-command",	OP_ENTER_COMMAND,		":" },
  { "next-line",	OP_NEXT_LINE,			M_ENTER_S },
  { "next-unread",	OP_MAIN_NEXT_UNREAD,		NULL },
  { "previous-new",	OP_MAIN_PREV_NEW,		NULL },
  { "previous-unread",	OP_MAIN_PREV_UNREAD,		NULL },
  { "half-up",		OP_PAGER_HALF_UP,		NULL },
  { "half-down",	OP_PAGER_HALF_DOWN,		NULL },
  { "previous-line",	OP_PREV_LINE,			NULL },
  { NULL,		0,				NULL }
};

struct binding_t OpAttach[] = {
  { "print-entry",	OP_PRINT,		"p" },
  { "save-entry",	OP_SAVE,		"s" },
  { "pipe-entry",	OP_PIPE,		"|" },
  { "view-mailcap",	OP_ATTACH_VIEW_MAILCAP,	"m" },
  { "reply",		OP_REPLY,               "r" },
  { "group-reply",	OP_GROUP_REPLY,         "g" },
  { "list-reply",	OP_LIST_REPLY,          "L" },
  { "tag-attachment",	OP_ATTACH_TAG,          "t" },
  { "view-text",	OP_ATTACH_VIEW_TEXT,    "T" },
  { NULL,		0,			NULL }
};

struct binding_t OpCompose[] = {
  { "attach-file",	OP_COMPOSE_ATTACH_FILE,		"a" },
  { "edit-bcc",		OP_COMPOSE_EDIT_BCC,		"b" },
  { "edit-cc",		OP_COMPOSE_EDIT_CC,		"c" },
  { "copy-file",        OP_SAVE,                        "C" },
  { "detach-file",	OP_DELETE,			"D" },
  { "edit-description",	OP_COMPOSE_EDIT_DESCRIPTION,	"d" },
  { "edit-message",	OP_COMPOSE_EDIT_MESSAGE,	"e" },
  { "edit-headers",	OP_COMPOSE_EDIT_HEADERS,	"E" },
  { "edit-file",	OP_COMPOSE_EDIT_FILE,		"\030e" },
  { "edit-encoding",	OP_COMPOSE_EDIT_ENCODING,	"\005" },
  { "edit-fcc",		OP_COMPOSE_EDIT_FCC,		"f" },
  { "ispell",		OP_COMPOSE_ISPELL,		"i" },
  { "print-attach",     OP_PRINT,                       "l" },
  { "redraw-screen",	OP_REDRAW,			"\014" },
  { "edit-mime",        OP_COMPOSE_EDIT_MIME,           "m" },
  { "new-mime",         OP_COMPOSE_NEW_MIME,            "n" },
  { "pgp-menu",		OP_COMPOSE_PGP_MENU,		"p" },
  { "postpone-message",	OP_COMPOSE_POSTPONE_MESSAGE,	"P" },
  { "exit",		OP_EXIT,			"q" },
  { "edit-reply-to",	OP_COMPOSE_EDIT_REPLY_TO,	"r" },
  { "rename-file",      OP_COMPOSE_RENAME_FILE,         "R" },
  { "edit-subject",	OP_COMPOSE_EDIT_SUBJECT,	"s" },
  { "edit-to",		OP_COMPOSE_EDIT_TO,		"t" },
  { "edit-type",	OP_COMPOSE_EDIT_TYPE,		"\024" },
  { "toggle-unlink",    OP_COMPOSE_TOGGLE_UNLINK,       "u" },
  { "view-attach",      OP_COMPOSE_VIEW_ATTACH,         "v" },
  { "send-message",	OP_COMPOSE_SEND_MESSAGE,	"y" },
  { "enter-command",	OP_ENTER_COMMAND,		":" },
  { "help",		OP_HELP,			"?" },
  { "pipe-entry",       OP_PIPE,                        "|" },
  { NULL,		0,				NULL }
};

struct binding_t OpPost[] = {
  { "delete-entry",	OP_DELETE,	"d" },
  { "undelete-entry",	OP_UNDELETE,	"u" },
  { NULL,		0,		NULL }
};

/* The file browser */
struct binding_t OpBrowser[] = {
  { "change-dir",	OP_CHANGE_DIRECTORY,	"c" },
  { "enter-mask",	OP_ENTER_MASK,		"m" },
  { "sort",		OP_SORT,		"o" },
  { "sort-reverse",	OP_SORT_REVERSE,	"O" },
  { "check-new",	OP_CHECK_NEW,		"\t" },
  { NULL,		0,			NULL }
};

struct binding_t OpEditor[] = {
  { "bol",		OP_EDITOR_BOL,			"\001" },
  { "backward-char",	OP_EDITOR_BACKWARD_CHAR,	"\002" },
  { "delete-char",	OP_EDITOR_DELETE_CHAR,		"\004" },
  { "eol",		OP_EDITOR_EOL,			"\005" },
  { "forward-char",	OP_EDITOR_FORWARD_CHAR,		"\006" },
  { "backspace",	OP_EDITOR_BACKSPACE,		"\010" },
  { "kill-eol",		OP_EDITOR_KILL_EOL,		"\013" },
  { "kill-line",	OP_EDITOR_KILL_LINE,		"\025" },
  { "kill-word",	OP_EDITOR_KILL_WORD,		"\027" },
  { "history-up",	OP_EDITOR_HISTORY_UP,		NULL },
  { "history-down",	OP_EDITOR_HISTORY_DOWN,		NULL },
  { NULL,		0,				NULL }
};
