/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"

#include <string.h>
#include <stdlib.h>

#define M_MODEFMT "-- Mutt: %s"

static void print_enriched_string (unsigned char *s, int do_color)
{
  while (*s)
  {
    if (*s < ctrl ('I'))
    {
      if (do_color)
	SETCOLOR (MT_COLOR_TREE);
      while (*s && *s < ctrl ('I'))
      {
	switch (*s)
	{
	  case 1:
	    addch (option (OPTASCIICHARS) ? '`' : ACS_LLCORNER);
	    break;
	  case 2:
	    addch (option (OPTASCIICHARS) ? ',' : ACS_ULCORNER);
	    break;
	  case 3:
	    addch (option (OPTASCIICHARS) ? '+' : ACS_LTEE);
	    break;
	  case 4:
	    addch (option (OPTASCIICHARS) ? '-' : ACS_HLINE);
	    break;
	  case 5:
	    addch (option (OPTASCIICHARS) ? '|' : ACS_VLINE);
	    break;
	  case 6:
	    addch (' ');
	    break;
	  case 7:
	    addch ('>');
	    break;
	  case 8:
	    addch ('*'); /* fake thread indicator */
	    break;
	}
	s++;
      }
      if (do_color)
	SETCOLOR (MT_COLOR_NORMAL);
    }
    else
    {
      addch (*s);
      s++;
    }
  }
}

void menu_pad_string (char *s, size_t l)
{
#if !defined(HAVE_BKGDSET) && !defined (USE_SLANG_CURSES)
  int n = strlen (s);
#endif
  int shift = option (OPTARROWCURSOR) ? 3 : 0;
  
  l--; /* save room for the terminal \0 */
  if (l > COLS - shift)
    l = COLS - shift;
#if !defined (HAVE_BKGDSET) && !defined (USE_SLANG_CURSES)
  /* we have to pad the string with blanks to the end of line */
  if (n < l)
  {
    while (n < l)
      s[n++] = ' ';
    s[n] = 0;
  }
  else
#endif
    s[l] = 0;
}

void menu_redraw_full (MUTTMENU *menu)
{
  SETCOLOR (MT_COLOR_NORMAL);
  clear ();

  if (option (OPTHELP))
  {
    SETCOLOR (MT_COLOR_STATUS);
    mvprintw (option (OPTSTATUSONTOP) ? LINES-2 : 0, 0, "%-*.*s", COLS, COLS, menu->help);
    SETCOLOR (MT_COLOR_NORMAL);
    menu->offset = 1;
    menu->pagelen = LINES - 3;
  }
  else
  {
    menu->offset = option (OPTSTATUSONTOP) ? 1 : 0;
    menu->pagelen = LINES - 2;
  }

  mutt_show_error ();

  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
}

void menu_redraw_index (MUTTMENU *menu)
{
  char buf[STRING];
  int i;

  for (i = menu->top; i < menu->top + menu->pagelen; i++)
  {
    if (i < menu->max)
    {
      menu->make_entry (buf, sizeof (buf), menu, i);
      menu_pad_string (buf, sizeof (buf));

      if (option (OPTARROWCURSOR))
      {
	CLEARLINE (i - menu->top + menu->offset);

	if (i == menu->current)
	{
	  SETCOLOR (MT_COLOR_INDICATOR);
	  addstr ("->");
	  SETCOLOR (MT_COLOR_NORMAL);
	  addch (' ');
	}
	else
	  move (i - menu->top + menu->offset, 3);
	print_enriched_string ((unsigned char *) buf, 1);
      }
      else
      {
	if (i == menu->current)
	{
	  SETCOLOR (MT_COLOR_INDICATOR);
	  BKGDSET (MT_COLOR_INDICATOR);
	}
	CLEARLINE (i - menu->top + menu->offset);
	print_enriched_string ((unsigned char *) buf, i != menu->current);
	if (i == menu->current)
	{
	  SETCOLOR (MT_COLOR_NORMAL);
	  BKGDSET (MT_COLOR_NORMAL);
	}
      }
    }
    else
      CLEARLINE (i - menu->top + menu->offset);
  }
  menu->redraw = 0;
}

void menu_redraw_motion (MUTTMENU *menu)
{
  char buf[STRING];
  
  move (menu->oldcurrent + menu->offset - menu->top, 0);

  if (option (OPTARROWCURSOR))
  {
    /* clear the pointer */
    addstr ("  ");

    if (menu->redraw & REDRAW_MOTION_RESYNCH)
    {
      clrtoeol ();
      menu->make_entry (buf, sizeof (buf), menu, menu->oldcurrent);
      menu_pad_string (buf, sizeof (buf));
      move (menu->oldcurrent + menu->offset - menu->top, 3);
      print_enriched_string ((unsigned char *) buf, 1);
    }

    /* now draw it in the new location */
    move (menu->current + menu->offset - menu->top, 0);
    SETCOLOR (MT_COLOR_INDICATOR);
    addstr ("->");
    SETCOLOR (MT_COLOR_NORMAL);
  }
  else
  {
    /* erase the current indicator */
    clrtoeol ();
    menu->make_entry (buf, sizeof (buf), menu, menu->oldcurrent);
    menu_pad_string (buf, sizeof (buf));
    print_enriched_string ((unsigned char *) buf, 1);

    /* now draw the new one to reflect the change */
    menu->make_entry (buf, sizeof (buf), menu, menu->current);
    menu_pad_string (buf, sizeof (buf));
    SETCOLOR (MT_COLOR_INDICATOR);
    BKGDSET (MT_COLOR_INDICATOR);
    CLEARLINE (menu->current - menu->top + menu->offset);
    print_enriched_string ((unsigned char *) buf, 0);
    SETCOLOR (MT_COLOR_NORMAL);
    BKGDSET (MT_COLOR_NORMAL);
  }
  menu->redraw = 0;
}

void menu_redraw_current (MUTTMENU *menu)
{
  char buf[STRING];
  
  move (menu->current + menu->offset - menu->top, 0);
  menu->make_entry (buf, sizeof (buf), menu, menu->current);
  menu_pad_string (buf, sizeof (buf));

  if (option (OPTARROWCURSOR))
  {
    clrtoeol ();
    SETCOLOR (MT_COLOR_INDICATOR);
    addstr ("->");
    SETCOLOR (MT_COLOR_NORMAL);
    addch (' ');
    menu_pad_string (buf, sizeof (buf));
    print_enriched_string ((unsigned char *) buf, 1);
  }
  else
  {
    SETCOLOR (MT_COLOR_INDICATOR);
    BKGDSET (MT_COLOR_INDICATOR);
    clrtoeol ();
    print_enriched_string ((unsigned char *) buf, 0);
    SETCOLOR (MT_COLOR_NORMAL);
    BKGDSET (MT_COLOR_NORMAL);
  }
  menu->redraw = 0;
}

void menu_jump (MUTTMENU *menu)
{
  int n;
  char buf[SHORT_STRING];

  mutt_ungetch (LastKey);
  buf[0] = 0;
  if (ci_get_field ("Jump to: ", buf, sizeof (buf), 0) == 0 && buf[0])
  {
    n = atoi (buf) - 1;
    if (n >= 0 && n < menu->max)
    {
      menu->oldcurrent = menu->current;
      menu->current = n;
      if (menu->current < menu->top)
      {
	while (menu->top > menu->current) menu->top -= menu->pagelen;
	if (menu->top < 0) menu->top = 0;
	menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      else if (menu->current >= menu->top + menu->pagelen)
      {
	do {
	  menu->top += menu->pagelen;
	} while (menu->top + menu->pagelen < menu->current);
	menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
      }
      else
	menu->redraw = REDRAW_MOTION;
    }
    else
      mutt_error ("Invalid index number.");
  }
}

void menu_next_line (MUTTMENU *menu)
{
  if (menu->top < menu->max - 1)
  {
    (menu->top)++;
    if (menu->current < menu->top)
      (menu->current)++;
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  }
  else
    mutt_error ("You cannot scroll down farther.");
}

void menu_prev_line (MUTTMENU *menu)
{
  if (menu->top > 0)
  {
    (menu->top)--;
    if (menu->current >= menu->top + menu->pagelen)
      (menu->current)--;
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  }
  else
    mutt_error ("You cannot scroll up farther.");
}

void menu_next_page (MUTTMENU *menu)
{
  if (menu->top + menu->pagelen <= menu->max - 1)
  {
    menu->top += menu->pagelen;
    menu->current = menu->top;
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  }
  else if (menu->current != menu->max - 1)
  {
    menu->current = menu->max - 1;
    menu->redraw = REDRAW_INDEX;
  }
  else
    mutt_error ("You are on the last page.");
}

void menu_prev_page (MUTTMENU *menu)
{
  if (menu->top > 0)
  {
    if ((menu->top -= menu->pagelen) < 0)
      menu->top = 0;
    if (menu->current > menu->top + menu->pagelen - 1)
      menu->current = menu->top + menu->pagelen - 1;
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  }
  else if (menu->current)
  {
    menu->current = menu->top;
    menu->redraw = REDRAW_INDEX;
  }
  else
    mutt_error("You are on the first page.");
}

void menu_top_page (MUTTMENU *menu)
{
  menu->oldcurrent = menu->current;
  menu->current = menu->top;
  menu->redraw = REDRAW_MOTION;
}

void menu_bottom_page (MUTTMENU *menu)
{
  menu->oldcurrent = menu->current;
  menu->current = menu->top + menu->pagelen - 1;
  if (menu->current > menu->max - 1) menu->current = menu->max - 1;
  menu->redraw = REDRAW_MOTION;
}

void menu_middle_page (MUTTMENU *menu)
{
  int i;

  menu->oldcurrent = menu->current;
  i = menu->top + menu->pagelen;
  if (i > menu->max - 1) i = menu->max - 1;
  menu->current = menu->top + (i - menu->top) / 2;
  menu->redraw = REDRAW_MOTION;
}

void menu_first_entry (MUTTMENU *menu)
{
  menu->oldcurrent = menu->current;
  if (menu->top > 0)
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  else
    menu->redraw = REDRAW_MOTION;
  menu->top = menu->current = 0;
}

void menu_last_entry (MUTTMENU *menu)
{
  menu->oldcurrent = menu->current;
  menu->current = menu->max - 1;
  if (menu->current > menu->top + menu->pagelen - 1)
  {
    while (menu->top + menu->pagelen  - 1 < menu->current)
      menu->top += menu->pagelen;
    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
  }
  else
    menu->redraw = REDRAW_MOTION;
}

MUTTMENU *mutt_new_menu (void)
{
  MUTTMENU *p = (MUTTMENU *) safe_malloc (sizeof (MUTTMENU));

  memset (p, 0, sizeof (MUTTMENU));
  p->offset = 1;
  p->redraw = REDRAW_FULL;
  p->pagelen = PAGELEN;
  return (p);
}

void mutt_menuDestroy (MUTTMENU **p)
{
  safe_free ((void **) &(*p)->searchBuf);
  safe_free ((void **) p);
}

#define M_SEARCH_UP   1
#define M_SEARCH_DOWN 2

static int menu_search (MUTTMENU *menu, int op)
{
  int r;
  int searchDir = (menu->searchDir == M_SEARCH_UP) ? -1 : 1;
  regex_t re;
  char buf[SHORT_STRING];

  if (op != OP_SEARCH_NEXT)
  {
    strfcpy (buf, menu->searchBuf ? menu->searchBuf : "", sizeof (buf));
    if (ci_get_field ("Search for: ", buf, sizeof (buf), M_CLEAR) != 0 || !buf[0])
      return (-1);
    safe_free ((void **) &menu->searchBuf);
    menu->searchBuf = safe_strdup (buf);
    menu->searchDir = (op == OP_SEARCH) ? M_SEARCH_DOWN : M_SEARCH_UP;
  }
  else if (!menu->searchBuf)
  {
    mutt_error ("No search pattern.");
    return (-1);
  }

  if ((r = REGCOMP (re, menu->searchBuf, mutt_which_case (menu->searchBuf))) != 0)
  {
    regerror (r, &re, buf, sizeof (buf));
    regfree (&re);
    mutt_error ("%s", buf);
    return (-1);
  }

  r = menu->current + searchDir;
  while (r >= 0 && r < menu->max)
  {
    if (menu->search (&re, r) == 0)
    {
      regfree (&re);
      return r;
    }

    r += searchDir;
  }

  regfree (&re);
  mutt_error ("Not found.");
  return (-1);
}

int mutt_menuLoop (MUTTMENU *menu)
{
  int i;
  char buf[STRING];

  FOREVER
  {
    curs_set (0);

    /* See if all or part of the screen needs to be updated.  */
    if (menu->redraw & REDRAW_FULL)
    {
      menu_redraw_full (menu);
      /* allow the caller to do any local configuration */
      return (OP_REDRAW);
    }

    if (menu->redraw & REDRAW_STATUS)
    {
      char tmp[STRING];
      int l;

      snprintf (buf, sizeof (buf), M_MODEFMT, menu->title);
      snprintf (tmp, sizeof (tmp), " -- %d/%d --", menu->top + 1, menu->max);
      l = strlen (tmp);

      SETCOLOR (MT_COLOR_STATUS);
      move (option (OPTSTATUSONTOP) ? 0 : LINES - 2, 0);
      printw ("%-*.*s", COLS - l, COLS - l, buf);
      addstr (tmp);
      SETCOLOR (MT_COLOR_NORMAL);
      menu->redraw &= ~REDRAW_STATUS;
    }

    if (menu->redraw & REDRAW_INDEX)
      menu_redraw_index (menu);
    else if (menu->redraw & (REDRAW_MOTION | REDRAW_MOTION_RESYNCH))
      menu_redraw_motion (menu);
    else if (menu->redraw == REDRAW_CURRENT)
      menu_redraw_current (menu);

    /* move the cursor out of the way */
    move (menu->current - menu->top + menu->offset,
	  (option (OPTARROWCURSOR) ? 2 : COLS-1));

    refresh ();
    i = km_dokey (menu->menu);
    curs_set (1);

#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
    if (Signals & S_SIGWINCH)
    {
      mutt_resize_screen ();
      menu->redraw = REDRAW_FULL;
      Signals &= ~S_SIGWINCH;
    }
#endif

    if (i == -1)
      continue;

    mutt_clear_error ();

    switch (i)
    {
      case OP_NEXT_ENTRY:

	if (menu->current < menu->max - 1)
	{
	  menu->oldcurrent = menu->current;
	  menu->current++;
	  if (menu->current >= menu->top + menu->pagelen)
	  {
	    menu->top = menu->current;
	    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
	  }
	  else
	    menu->redraw = REDRAW_MOTION;
	}
	else
	  mutt_error ("You are on the last entry.");
	break;

      case OP_PREV_ENTRY:

	if (menu->current > 0)
	{
	  menu->oldcurrent = menu->current;
	  if (--menu->current < menu->top)
	  {
	    if ((menu->top = menu->current - menu->pagelen + 1) < 0)
	      menu->top = 0;
	    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
	  }
	  else
	    menu->redraw = REDRAW_MOTION;
	}
	else
	  mutt_error("You are on the first entry.");
	break;

      case OP_NEXT_PAGE:

	menu_next_page (menu);
	break;

      case OP_PREV_PAGE:

	menu_prev_page (menu);
	break;
	
      case OP_NEXT_LINE:
	
	menu_next_line (menu);
	break;
	
      case OP_PREV_LINE:
	
	menu_prev_line (menu);
	break;

      case OP_FIRST_ENTRY:

	menu_first_entry (menu);
	break;

      case OP_LAST_ENTRY:

	menu_last_entry (menu);
	break;

      case OP_TOP_PAGE:

	menu_top_page (menu);
	break;

      case OP_MIDDLE_PAGE:

	menu_middle_page (menu);
	break;

      case OP_BOTTOM_PAGE:

	menu_bottom_page (menu);
	break;

      case OP_SEARCH:
      case OP_SEARCH_REVERSE:
      case OP_SEARCH_NEXT:
	
	if (menu->search)
	{
	  menu->oldcurrent = menu->current;
	  if ((menu->current = menu_search (menu, i)) != -1)
	  {
	    if (menu->top + menu->pagelen - 1 < menu->current)
	    {
	      while (menu->top + menu->pagelen - 1 < menu->current)
		menu->top += menu->pagelen;
	      menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
	    }
	    else if (menu->current < menu->top)
	    {
	      while (menu->top > menu->current) menu->top -= menu->pagelen;
	      menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
	    }
	    else
	      menu->redraw = REDRAW_MOTION;
	  }
	  else
	    menu->current = menu->oldcurrent;
	}
	else
	  mutt_error ("Search is not implemented for this menu.");
	break;

      case OP_JUMP:
	
	menu_jump (menu);
	break;

      case OP_ENTER_COMMAND:

	mutt_enter_command ();
	break;

      case OP_REDRAW:

	clearok (stdscr, TRUE);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_HELP:

	mutt_help (menu->menu);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_NULL:

	km_error_key (menu->menu);
	break;

      default:
	return (i);
    }
  }
  /* not reached */
}
