/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "mutt.h"
#include "mutt_curses.h"
#include "rfc2047.h"

#include <sys/wait.h>
#include <string.h>

/* See muttlib.h for the flags to pass to this function. */
int mutt_copy_header (FILE *in, HEADER *h, FILE *out, int flags)
{
  char buf[HUGE_STRING];
  int ignore = 0, from = 0;

  buf[sizeof (buf) - 1] = 0;

  if (ftell (in) != h->offset)
    fseek (in, h->offset, 0);

  while (ftell (in) < h->content->offset)
  {
    if (fgets (buf, sizeof (buf) - 1, in) == NULL)
      break;

    if (!from && strncmp ("From ", buf, 5) == 0)
    {
      if ((flags & CH_FROM) == 0)
	continue;
      from = 1;
    }
    else if (buf[0] == '\n')
      break; /* end of header */

    if (flags & CH_WEED)
    {
      if (buf[0] == ' ' || buf[0] == '\t')
      {
	if (ignore)
	  continue;
      }
      else if (mutt_matches_ignore (buf, Ignore) &&
	       !mutt_matches_ignore (buf, UnIgnore))
	ignore = 1;
      else
	ignore = 0;
    }
    
    if (!ignore)
    {
      if ((flags & (CH_UPDATE | CH_XMIT | CH_NOSTATUS)) &&
	  (strncasecmp ("Status:", buf, 7) == 0 ||
	   strncasecmp ("X-Status:", buf, 9) == 0))
	continue;
      if ((flags & (CH_UPDATE | CH_XMIT)) &&
	  (strncasecmp ("Content-Length:", buf, 15) == 0 ||
	   strncasecmp ("Lines:", buf, 6) == 0))
	continue;

      if (flags & CH_DECODE)
	rfc2047_decode (buf, buf, sizeof (buf));

      if (flags & CH_PREFIX)
      {
	if (fputs (Prefix, out) == EOF)
	  return (-1);
      }

      if (fputs (buf, out) == EOF)
	return (-1);
    }
  }

  if (flags & CH_UPDATE)
  {
    if ((flags & CH_NOSTATUS) == 0)
    {
      if (h->old || h->read)
      {
	if (fputs ("Status: ", out) == EOF)
	  return (-1);

	if (h->read)
	{
	  if (fputs ("RO", out) == EOF)
	    return (-1);
	}
	else if (h->old)
	{
	  if (fputc ('O', out) == EOF)
	    return (-1);
	}

	if (fputc ('\n', out) == EOF)
	  return (-1);
      }

      if (h->flagged || h->replied)
      {
	if (fputs ("X-Status: ", out) == EOF)
	  return (-1);

	if (h->replied)
	{
	  if (fputc ('A', out) == EOF)
	    return (-1);
	}

	if (h->flagged)
	{
	  if (fputc ('F', out) == EOF)
	    return (-1);
	}
	
	if (fputc ('\n', out) == EOF)
	  return (-1);
      }
    }

    fprintf (out, "Content-Length: %ld\n", h->content->length);
    fprintf (out, "Lines: %d\n", h->lines);
  }

  if ((flags & CH_XMIT) == 0)
  {
    if (fputc ('\n', out) == EOF) /* add header terminator */
      return (-1);
  }
  
  return 0;
}

static void print_msg (HEADER *h, STATE *s)
{
  MESSAGE *msg;

  mutt_parse_mime_message (h);
  

  if ((msg = mx_open_message (Context, h->msgno)))
  {
    mutt_copy_header (msg->fp, h, s->fpout, CH_WEED | CH_DECODE);
    s->fpin = msg->fp;
    mutt_body_handler (h->content, s);

    /* add a formfeed */
    state_putc ('\f', s);

    s->fpin = NULL;
    mx_close_message (&msg);
  }
}

void mutt_print_message (HEADER *h)
{
  STATE s;
  int i, count = 0;
  pid_t thepid;

  if (query_quadoption (OPT_PRINT,
			h ? "Print message?" : "Print tagged messages?") != M_YES)
    return;
  memset (&s, 0, sizeof (STATE));
  mutt_endwin (NULL);
  if ((thepid = mutt_create_filter (PrintCmd, &s.fpout, NULL, NULL)) == -1)
    return;
  if (h)
  {
    print_msg (h, &s);
    count++;
  }
  else
  {
    for (i = 0 ; i < Context->vcount ; i++)
    {
      if (Context->hdrs[Context->v2r[i]]->tagged)
      {
	print_msg (Context->hdrs[Context->v2r[i]], &s);
	count++;
      }
    }
  }
  fclose (s.fpout);
  mutt_wait_filter (thepid);
  mutt_message ("Message%s printed", (count > 1) ? "s" : "");
}
