/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "rfc1524.h"


#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>

struct attachptr
{
  BODY *content;
  struct attachptr *next;
  char *tree;
  int level;
} **idx = NULL;

static struct attachptr *gen_attach_list (BODY *m, int level)
{
  struct attachptr *top = NULL, *end = NULL, *new;

  while (m)
  {
    if (m->type == TYPEMULTIPART && m->parts)
    {
      if (top)
	end->next = gen_attach_list (m->parts, level);
      else
	end = top = gen_attach_list (m->parts, level);
      while (end->next)
	end = end->next;
    }
    else
    {
      new = (struct attachptr *) safe_malloc (sizeof (struct attachptr));
      new->next = NULL;
      new->content = m;
      new->level = level;

      if (top)
      {
        end->next = new;
        end = new;
      }
      else
        top = end = new;

      if (m->type == TYPEMESSAGE &&
	  (!strcasecmp (m->subtype, "rfc822") || !strcasecmp (m->subtype, "news")) &&
	  is_multipart (m->parts))
      {
	end->next = gen_attach_list (m->parts, level + 1);
	while (end->next)
	  end = end->next;
      }
    }

    m = m->next;
  }

  return (top);
}

void mutt_display_attachment (FILE *fp, BODY *a, int flag)
{
  char tempfile[_POSIX_PATH_MAX] = "";
  char descrip[SHORT_STRING];

  if (flag == M_MAILCAP || (flag == M_REGULAR && mutt_needs_mailcap (a)))
  {
    char type[STRING];
    char command[STRING];
    rfc1524_entry *entry = rfc1524_new_entry ();

    snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
    if (rfc1524_mailcap_lookup (a, type, entry, 0))
    {
      if (entry->command)
      {
        rfc1524_expand_filename(entry->nametemplate,a->filename,
	                        tempfile,sizeof(tempfile));
	strfcpy (command, entry->command, sizeof (command));
	if (!rfc1524_expand_command(a,tempfile, type, command, sizeof(command)))
	{
	  /* rfc1524_expand_command returns 0 if the file is required */
	  mutt_save_attachment (fp, a, tempfile);
	  if (entry->copiousoutput)
	  {
	    /* Use a pager: The assumption is the info is coming from 
	     * stdout of the command without a pager, so we use our own
	     */
	    char outfile[_POSIX_PATH_MAX] = "";
	    FILE *fpin;
	    FILE *tfp;

	    mutt_mktemp (outfile);
	    if ((tfp = safe_fopen (outfile, "w")) == NULL)
	      mutt_perror (outfile);
	    else
	    {
	      pid_t thepid = mutt_create_filter (command, NULL, &fpin, NULL);

	      mutt_endwin (NULL);
	      mutt_copy_stream (fpin, tfp);
	      mutt_wait_filter (thepid);
	      fclose (fpin);
	      fclose (tfp);

	      if (a->description)
		snprintf (descrip, sizeof(descrip),
			  "---Command: %-20.20s Description: %s",command,a->description);
	      else
		snprintf (descrip, sizeof (descrip),
			  "---Command: %-30.30s Attachment: %s/%s", command, 
			  TYPE (a->type), a->subtype);
	      mutt_do_pager (descrip, outfile, 0);
	    }
	  }
	  else
	  {
	    mutt_endwin (NULL);
	    mutt_system (command);
	  }
	}
	else
	{
	  /* rfc1524_expand_command returns 1 if it takes a pipe */
	  if (entry->copiousoutput)
	  {
	    mutt_pipe_attachment (fp, a, command, tempfile);
	    if (a->description)
	      snprintf (descrip, sizeof(descrip),
			"---Command: %-20.20s Description: %s",
			command, a->description);
	    else
	      snprintf (descrip, sizeof (descrip),
			"---Command: %-30.30s Attachment: %s/%s", command,
			TYPE (a->type), a->subtype);
	    mutt_do_pager (descrip, tempfile, 0);
	  }
	  else
	    mutt_pipe_attachment (fp, a, command, NULL);
	}
      }
    }
    else
    {
      /* no mailcap type defined */
    }

    rfc1524_free_entry (&entry);
    mutt_unlink (tempfile);
  }
  else
  {
    STATE s;

    if (a->filename)
      strfcpy (tempfile, a->filename, sizeof (tempfile));
    mutt_adv_mktemp (tempfile);

    memset (&s, 0, sizeof (s));
    if ((s.fpout = safe_fopen (tempfile, "w")) == NULL)
    {
      mutt_perror (tempfile);
      return;
    }

    s.fpin = fp;

    if (flag == M_AS_TEXT)
    {
      /* don't try to do any fancy handling, just let me see the raw data! */
      mutt_decode_attachment (a, &s);
    }
    else
    {
      s.displaying = 1;
      mutt_body_handler (a, &s);
    }
    fclose (s.fpout);

    if (a->description)
      strfcpy (descrip, a->description, sizeof (descrip));
    else
      snprintf (descrip, sizeof (descrip),
		"---Attachment: %s/%s", TYPE (a->type), a->subtype);
    mutt_do_pager (descrip, tempfile,
		   (a->type == TYPEMESSAGE && (!strcasecmp ("rfc822", a->subtype) || !strcasecmp ("news", a->subtype))));
  }
}

void attach_entry (char *b, size_t blen, MUTTMENU *menu, int num)
{
  char t[SHORT_STRING];
  char size[SHORT_STRING];
  BODY *m = idx[num]->content;

  mutt_pretty_size (size, sizeof (size), m->length);
  snprintf (t, sizeof (t), "[%.7s/%.10s, %.6s, %s]",
	    TYPE (m->type), m->subtype, ENCODING (m->encoding), size);
  snprintf (b, blen, " %c %2d %-33.33s %s%s",
            m->untagged ? ' ' : '*',
	    num + 1,
	    t,
	    idx[num]->tree ? idx[num]->tree : "",
	    m->description ? m->description : (m->filename ? m->filename : "<no description>"));
}

void mutt_view_attachments (HEADER *hdr)
{
  struct attachptr *list, *plist;
  int idxlen = 0, count = 0;
  char buf[STRING];
  char helpstr[SHORT_STRING];
  char *s;
  MUTTMENU *menu;
  BODY *cur;
  FILE *fp;
  MESSAGE *msg;
  
  /* make sure we have parsed this message */
  mutt_parse_mime_message (hdr);

  if ((msg = mx_open_message (Context, hdr->msgno)) == NULL)
    return;

  {
    fp = msg->fp;
    cur = hdr->content;
  }

  plist = list = gen_attach_list (cur, 0);
  idx = safe_malloc (sizeof (struct attachptr *) * ((idxlen = 5)));
  idxlen = 5;

  for (count = 0; plist; count++, plist = plist->next)
  {
    if (count == idxlen)
      safe_realloc ((void **)&idx, ((idxlen += 5)) * sizeof (struct attachptr *));
    idx[count] = plist;
    if (plist->level)
    {
      s = buf + 2 * (plist->level - 1);
      *s++ = (plist->content->next) ? '\003' : '\001';
      *s++ = '\004';
      *s++ = '\007';
    }
    else
      s = buf;
    *s = 0;
    plist->tree = safe_strdup (buf);

    if (plist->level)
    {
      s = buf + 2 * (plist->level - 1);
      *s++ = (plist->content->next) ? '\005' : '\006';
      *s++ = '\006';
    }
  }

  menu = mutt_new_menu ();
  menu->max = count;
  menu->make_entry = attach_entry;
  menu->menu = MENU_ATTACH;
  menu->title = "Attachments";

  helpstr[0] = 0;
  mutt_make_help (buf, sizeof (buf), "Exit  ", MENU_ATTACH, OP_EXIT);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Save  ", MENU_ATTACH, OP_SAVE);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Pipe  ", MENU_ATTACH, OP_PIPE);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Print  ", MENU_ATTACH, OP_PRINT);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Help", MENU_ATTACH, OP_HELP);
  strcat (helpstr, buf);
  menu->help = helpstr;

  FOREVER
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_GENERIC_SELECT_ENTRY:

	mutt_display_attachment (fp, idx[menu->current]->content, M_REGULAR);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_ATTACH_VIEW_MAILCAP:

	mutt_display_attachment (fp, idx[menu->current]->content,  M_MAILCAP);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_ATTACH_VIEW_TEXT:

	mutt_display_attachment (fp, idx[menu->current]->content,  M_AS_TEXT);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_PRINT:

	if (query_quadoption(OPT_PRINT, "Print Attachment?") == M_YES)
	  if (mutt_print_attachment (fp, idx[menu->current]->content))
	    menu->redraw = REDRAW_FULL;
	break;

      case OP_PIPE:

	buf[0] = 0;
	if (ci_get_field ("Pipe to: ", buf, sizeof (buf), 0) == 0 && buf[0])
	{
	  mutt_expand_path (buf, sizeof (buf));
	  mutt_pipe_attachment (fp, idx[menu->current]->content, buf, NULL);
	}
	break;

      case OP_SAVE:

	if (idx[menu->current]->content->filename)
	  strfcpy (buf, idx[menu->current]->content->filename, sizeof (buf));
	else
	  buf[0] = 0;
	if (ci_get_field ("Save to file: ", buf, sizeof (buf), M_FILE | M_CLEAR) != 0 ||
	    !buf[0])
	  break;
	mutt_expand_path (buf, sizeof (buf));
	if (mutt_save_attachment (fp, idx[menu->current]->content, buf) == 0)
	  mutt_message ("Attachment saved.");
	break;

      case OP_ATTACH_TAG:

	idx[menu->current]->content->untagged =
	  idx[menu->current]->content->untagged ? 0 : 1;
	menu->redraw = REDRAW_CURRENT;
	break;

      case OP_REPLY:

	ci_send_message (SENDREPLY, NULL, NULL, hdr);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_GROUP_REPLY:

	ci_send_message (SENDREPLY|SENDGROUPREPLY, NULL, NULL, hdr);
	menu->redraw = REDRAW_FULL;
	break;

      case OP_LIST_REPLY:

	ci_send_message (SENDREPLY|SENDLISTREPLY, NULL, NULL, hdr);
	menu->redraw = REDRAW_FULL;
	break;

	/* This doesn't work yet because of how mutt handles forwards */
	/*      case OP_FORWARD_MESSAGE:

		ci_send_message (SENDFORWARD, NULL, NULL, hdr);
		menu->redraw = REDRAW_FULL;
		break;
	 */

      case OP_EXIT:

	mx_close_message (&msg);

	while (count-- > 0)
	{
	  idx[count]->content->untagged = 0;
	  safe_free ((void **) &idx[count]->tree);
	  safe_free ((void **) &idx[count]);
	}
	safe_free ((void **) &idx);


	mutt_menuDestroy  (&menu);
	return;
    }
  }

  /* not reached */
}
